#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <termios.h>

#define DEFAULT_DEVICE         "/dev/ttyS0"
#define FULL_PACKET_SIZE       5
#define START_CALIBRATION      'K'
#define CALIBRATION_STATUS_1  0xC1
#define CALIBRATION_STATUS_2  0xC2
#define CALIBRATION_ERROR     0xC0

void print_packet_info( unsigned char *packet)
{
   int abs_x;
   int abs_y;

   if ( ! ( packet[0] & 0x40 ) )
     {
	
	fprintf( stdout, "%3d %3d %3d %3d %3d  >> ", packet[0], packet[1], packet[2], packet[3], packet[4]);
	
	abs_y = (((unsigned short)packet[0]) << 7 ) & 0x0F80 ;
	abs_x = (((unsigned short)packet[1]) << 7 ) & 0x0F80 ;
	
	abs_y |= ((unsigned short)packet[2]) & 0x007F; 
	abs_x |= ((unsigned short)packet[3]) & 0x007F;
	
	fprintf( stdout, "abs_x = %5d   |   abs_y = %5d\n", abs_x, abs_y);
     }
   else
     {
	fprintf( stderr, "LIFT OFF\n");
     }
   
   
}

void do_calibration( FILE *f_dev)
{
   unsigned char start_calibration = START_CALIBRATION;
   unsigned char a_byte;
   int done = 0;
   
   fwrite( &start_calibration, sizeof(unsigned char), 1, f_dev);
   
   fprintf( stdout, "Please, push the first alignment point\n");
   
   while( !done)
     {
	fread( &a_byte, sizeof(unsigned char), 1, f_dev);
	
	if ( a_byte == CALIBRATION_STATUS_1)
	  {
	     done = 1;
	  }
	else if ( a_byte == CALIBRATION_ERROR)
	  {
	     fprintf( stderr, "Calibration error ! \n");
	     exit( EXIT_FAILURE);
	  }
	else
	  {     
	     fprintf( stderr, "%3d ", a_byte);
	     fflush( stderr);
	  }
     }

   fprintf( stdout, "Please, push the second alignment point\n");

   while( !done)
     {
	fread( &a_byte, sizeof(unsigned char), 1, f_dev);
	
	if ( a_byte == CALIBRATION_STATUS_2)
	  {
	     done = 1;
	  }
	else if ( a_byte == CALIBRATION_ERROR)
	  {
	     fprintf( stderr, "Calibration error ! \n");
	     exit( EXIT_FAILURE);
	  }
	else
	  {     
	     //fprintf( stderr, ".");
	     //fflush( stderr);
	  }
     }

   fprintf( stdout, "Calibration done.\n");
   

}

void usage( char *cmd)
{
   fprintf( stderr, "usage: %s [options]\n\n", cmd);
   fprintf( stderr, "Options: -calibrate            Start calibration\n");
   fprintf( stderr, "         -dev <serial_device>  Use this device (default %s)\n", DEFAULT_DEVICE);
}

	   
int main( int argc, char **argv)
{
   int arg;
   int calibrate = 0;
   FILE *f_dev;
   int fd_dev;
   struct termios cfg;
   char* device_name = DEFAULT_DEVICE;
   unsigned char a_byte;
   unsigned char last_packet[FULL_PACKET_SIZE];
   unsigned char partial_packet[FULL_PACKET_SIZE];
   unsigned char current_packet[FULL_PACKET_SIZE];
   int synchronized = 0;
   int reading_a_packet = 0;
   int partial_packet_position = 0;
   
 
   for ( arg = 1 ; arg < argc ; arg++)
     {
	if ( strcmp( "-calibrate", argv[arg]) == 0 )
	  {
	     calibrate = 1;
	  }
	else
	  {
	     usage( argv[0]);
	     exit( EXIT_FAILURE);
	  }
     }
     
   
   // 1) Open serial device
  
   f_dev = fopen( device_name, "rw");
   
   if ( f_dev == NULL )
     {
	fprintf( stderr, "%s: can't open device %s\n", argv[0], device_name);
	exit (EXIT_FAILURE);
     }
   
   // 2) Configure Serial Port to 9600 bps / 1 start bit / 8 data bits / 1 stop bit / no parity   
   
   fd_dev = fileno( f_dev);
   
   bzero( &cfg, sizeof(cfg));
   cfg.c_cflag = CRTSCTS | CS8 | /*CSTOPB |*/ CLOCAL | CREAD;
   cfg.c_iflag = IGNPAR;
   cfg.c_oflag = IGNPAR;
   //cfg.c_lflag = ICANON;
   cfsetispeed( &cfg, B9600);
   cfsetospeed( &cfg, B9600);
   
   tcsetattr( fd_dev, TCSANOW, &cfg);
      
   // 3) If request calibrate
   
   if ( calibrate)
     do_calibration( f_dev);
   
   // 4) Read data

   while( 1)
     {
	if ( fread( &a_byte, sizeof( unsigned char), 1, f_dev) == 1 )
	  {
	     if ( synchronized)
	       {
		  partial_packet[partial_packet_position++] = a_byte;
		  
		  if ( a_byte & 0x80 ) // if the most significant bit is on 
		    {
		       // this is the last byte of the packet
		       int b;
		       
		       memcpy( current_packet, last_packet, FULL_PACKET_SIZE);
		       
		       for( b = 0 ; b < partial_packet_position ; b++) 
			 current_packet[b+(FULL_PACKET_SIZE-partial_packet_position)] = partial_packet[b];
		       
		       fprintf( stdout, "[%d] ", partial_packet_position);
		       
		       partial_packet_position = 0;
		       
		       print_packet_info( current_packet);
		    
		       memcpy( last_packet, current_packet, FULL_PACKET_SIZE);
		    }
		  
		  if ( partial_packet_position >= FULL_PACKET_SIZE )
		    {
		       fprintf( stderr, "%s: [ERROR] we read a packet larger than %d bytes\n", argv[0], FULL_PACKET_SIZE);
		       fclose( f_dev);
		       exit (EXIT_FAILURE);
		    }
	       }
	     else // not synchronized
	       {
		  // to be synchronized we need to read a full packet

		  if ( reading_a_packet)
		    {
		       
		       partial_packet[partial_packet_position++] = a_byte;
		       
		       if ( ( a_byte & 0x80 ) && partial_packet_position == FULL_PACKET_SIZE)
			 {
			    // Yes !
			    fprintf( stderr, "%s: SYCHRONIZED !\n", argv[0]);
			    partial_packet_position = 0;
			    synchronized = 1;
			 }
		       else if ( a_byte & 0x80 )
			 {
			    partial_packet_position = 0;
			    fprintf( stderr, "%s: not synchronized...\n", argv[0]);
			 }
		       else if ( partial_packet_position >= FULL_PACKET_SIZE)
			 {
			    fprintf( stderr, "%s: [ERROR] we read a packet larger than %d bytes\n", argv[0], FULL_PACKET_SIZE);
			    fclose (f_dev);
			    exit( EXIT_FAILURE);
			 }   
		    }
		  else
		    {
		       if ( a_byte & 0x80 )
			 {
			    reading_a_packet = 1;
			    partial_packet_position = 0;
			    fprintf( stderr, "msb on\n");
			 }
		       else
			 {
			    if ( isalnum( a_byte))
			      fprintf( stderr, "%c", a_byte);
			    else
			      fprintf( stderr, ".");
			    fflush( stderr);
			 }
		       
		    }
	       }
	  }
     }
   
   exit( EXIT_SUCCESS);
   
}
